use std::io::{Cursor, Read, Seek, SeekFrom};
use std::path::{Path, PathBuf};

use bitflags::bitflags;
use color_eyre::eyre::Context;
use color_eyre::{eyre, SectionExt};
use color_eyre::{Help, Result};
use oodle::{OodleLZ_CheckCRC, OodleLZ_FuzzSafe};
use serde::{Deserialize, Serialize};
use tokio::fs;

use crate::binary::sync::{ReadExt, WriteExt};
use crate::bundle::file::UserFile;
use crate::murmur::{HashGroup, IdString32, IdString64};
use crate::{BundleFile, BundleFileType, BundleFileVariant};

#[derive(Clone, Debug, Deserialize, Serialize)]
struct TextureDefinition {
    common: TextureDefinitionPlatform,
    // Stingray supports per-platform sections here, where you can create overrides with the same
    // values as in `common`. But since we only support PC, we don't need to implement
    // that.
}

#[derive(Clone, Debug, Deserialize, Serialize)]
struct TextureDefinitionPlatform {
    input: TextureDefinitionInput,
    output: TextureDefinitionOutput,
}

#[derive(Clone, Debug, Deserialize, Serialize)]
struct TextureDefinitionInput {
    filename: String,
}

#[derive(Clone, Debug, Deserialize, Serialize)]
struct TextureDefinitionOutput {
    category: String,
}

bitflags! {
    #[derive(Clone, Copy, Debug)]
    struct TextureFlags: u32 {
        const STREAMABLE = 0b0000_0001;
        const UNKNOWN = 1 << 1;
        const SRGB = 1 << 8;
    }
}

#[derive(Clone, Debug)]
struct TextureHeader {
    flags: TextureFlags,
    n_streamable_mipmaps: u32,
    width: u32,
    height: u32,
}

impl TextureHeader {
    #[tracing::instrument(skip(r))]
    fn from_binary(mut r: impl ReadExt) -> Result<Self> {
        let flags = r.read_u32().and_then(|bits| {
            TextureFlags::from_bits(bits)
                .ok_or_else(|| eyre::eyre!("Unknown bits set in TextureFlags: {:032b}", bits))
        })?;
        let n_streamable_mipmaps = r.read_u32()?;
        let width = r.read_u32()?;
        let height = r.read_u32()?;

        // Don't quite know yet what this is, only that it is related to mipmaps.
        // The reference to "streamable mipmaps" comes from VT2, so far.
        // As such, it might be related to the stream file, but since all texture files have it,
        // The engine calculates some offset and then moves 68 bytes at that offset to the beginning.
        // Hence the split between `68` and `60` in the length.
        r.seek(SeekFrom::Current(68 + 60))?;

        Ok(Self {
            flags,
            n_streamable_mipmaps,
            width,
            height,
        })
    }

    #[tracing::instrument(skip(w))]
    fn to_binary(&self, mut w: impl WriteExt) -> Result<()> {
        eyre::ensure!(
            self.flags.is_empty() && self.n_streamable_mipmaps == 0,
            "Only textures are supported where `flags == 0` and `n_streamable_mipmaps == 0`."
        );

        w.write_u32(self.flags.bits())?;
        w.write_u32(self.n_streamable_mipmaps)?;
        w.write_u32(self.width)?;
        w.write_u32(self.height)?;

        // See `from_binary` about this unknown section.
        let buf = [0; 148];
        w.write_all(&buf)?;

        Ok(())
    }
}

#[derive(Clone, Debug)]
struct Texture {
    header: TextureHeader,
    data: Vec<u8>,
    stream: Option<Vec<u8>>,
    category: IdString32,
}

impl Texture {
    #[tracing::instrument(skip(ctx, r, stream_r))]
    fn from_binary(
        ctx: &crate::Context,
        mut r: impl Read + Seek,
        mut stream_r: Option<impl Read>,
    ) -> Result<Self> {
        // Looking at the executable in IDA, there is one other valid value: `2`.
        // If this ever comes up in the game data, I'll have to reverse engineer the
        // (de)compression algorithm through IDA.
        let compression_type = r.read_u32()?;
        eyre::ensure!(
            compression_type == 1,
            "Unknown compression type for texture '{}'",
            compression_type
        );

        let compressed_size = r.read_u32()? as usize;
        let uncompressed_size = r.read_u32()? as usize;

        let out_buf = {
            let mut comp_buf = vec![0; compressed_size];
            r.read_exact(&mut comp_buf)?;

            oodle::decompress(
                comp_buf,
                uncompressed_size,
                OodleLZ_FuzzSafe::No,
                OodleLZ_CheckCRC::No,
            )?
        };

        eyre::ensure!(
            out_buf.len() == uncompressed_size,
            "Length of decompressed buffer did not match expected value. Expected {}, got {}",
            uncompressed_size,
            out_buf.len()
        );

        // No idea what this number is supposed to mean.
        // Even the game engine just skips this one.
        r.skip_u32(0x43)?;

        let header = TextureHeader::from_binary(&mut r)?;

        let meta_size = r.read_u32()?;

        eyre::ensure!(
            meta_size == 0 || stream_r.is_some(),
            "Compression chunks and stream file don't match up. meta_size = {}, stream = {}",
            meta_size,
            stream_r.is_some()
        );

        let stream = if let Some(stream_r) = stream_r.as_mut() {
            // Number of compression chunks in the stream file
            let num_chunks = r.read_u32()?;
            r.skip_u16(0)?;

            {
                let num_chunks_1 = r.read_u16()? as u32;

                eyre::ensure!(
                    num_chunks == num_chunks_1,
                    "Chunk numbers don't match. first = {}, second = {}",
                    num_chunks,
                    num_chunks_1
                );
            }

            const RAW_SIZE: usize = 0x10000;
            let mut stream_raw = Vec::new();
            let mut last = 0;

            for i in 0..num_chunks {
                let offset_next = r.read_u32()? as usize;
                let size = offset_next - last;

                let span = tracing::info_span!(
                    "read stream chunk",
                    num_chunks,
                    i,
                    chunk_size = size,
                    offset = last
                );
                let _enter = span.enter();

                let mut buf = vec![0; size];
                stream_r
                    .read_exact(&mut buf)
                    .wrap_err("Failed to read chunk from stream file")?;

                let raw =
                    oodle::decompress(&buf, RAW_SIZE, OodleLZ_FuzzSafe::No, OodleLZ_CheckCRC::No)
                        .wrap_err("Failed to decompress stream chunk")?;

                stream_raw.extend_from_slice(&raw);

                last = offset_next;
            }

            Some(stream_raw)
        } else {
            None
        };

        let category = ctx.lookup_hash_short(r.read_u32()?, HashGroup::TextureCategory);

        Ok(Self {
            category,
            header,
            data: out_buf,
            stream,
        })
    }

    #[tracing::instrument(skip(w))]
    fn to_binary(&self, mut w: impl WriteExt) -> Result<()> {
        let compression_type = 1;
        w.write_u32(compression_type)?;

        let comp_buf = oodle::compress(&self.data).wrap_err("Failed to compress DDS data")?;

        w.write_u32(comp_buf.len() as u32)?;
        w.write_u32(self.data.len() as u32)?;
        w.write_all(&comp_buf)?;

        // Unknown field, which the engine seems to ignore.
        // All game files have the same value here, so we just mirror that.
        w.write_u32(0x43)?;

        self.header.to_binary(&mut w)?;

        // More data not fully figured out, yet.
        let meta_size = 0;
        w.write_u32(meta_size)?;

        w.write_u32(self.category.to_murmur32().into())?;
        Ok(())
    }

    #[tracing::instrument]
    fn to_sjson(&self, filename: String) -> Result<String> {
        let texture = TextureDefinition {
            common: TextureDefinitionPlatform {
                input: TextureDefinitionInput { filename },
                output: TextureDefinitionOutput {
                    category: self.category.display().to_string(),
                },
            },
        };
        serde_sjson::to_string(&texture).wrap_err("Failed to serialize texture definition")
    }

    #[tracing::instrument]
    fn to_user_files(&self, name: String) -> Result<Vec<UserFile>> {
        let mut files = Vec::with_capacity(3);

        // TODO: Don't clone.

        if let Some(stream) = &self.stream {
            let stream_name = PathBuf::from(&name).with_extension("stream");
            files.push(UserFile::with_name(
                stream.clone(),
                stream_name.display().to_string(),
            ));
        }

        {
            let data = self.to_sjson(name.clone())?.as_bytes().to_vec();
            let name = PathBuf::from(&name)
                .with_extension("texture")
                .display()
                .to_string();
            files.push(UserFile::with_name(data, name));
        }

        files.push(UserFile::with_name(self.data.clone(), name));
        Ok(files)
    }
}

#[tracing::instrument(skip(ctx, data), fields(buf_len = data.as_ref().len()))]
pub(crate) async fn decompile_data(
    ctx: &crate::Context,
    name: String,
    data: impl AsRef<[u8]>,
    stream_file_name: Option<PathBuf>,
) -> Result<Vec<UserFile>> {
    let mut r = Cursor::new(data.as_ref());
    let mut stream_r = if let Some(file_name) = stream_file_name {
        let stream_data = fs::read(&file_name)
            .await
            .wrap_err_with(|| format!("Failed to read stream file '{}'", file_name.display()))?;
        Some(Cursor::new(stream_data))
    } else {
        None
    };

    let texture = Texture::from_binary(ctx, &mut r, stream_r.as_mut())?;
    texture
        .to_user_files(name)
        .wrap_err("Failed to build user files")
}

#[tracing::instrument(skip(ctx))]
pub(crate) async fn decompile(
    ctx: &crate::Context,
    name: String,
    variant: &BundleFileVariant,
) -> Result<Vec<UserFile>> {
    if !variant.external() {
        tracing::debug!("Decompiling texture from bundle data");

        let stream_file_name = variant.data_file_name().map(|name| match &ctx.game_dir {
            Some(dir) => dir.join("bundle").join(name),
            None => PathBuf::from("bundle").join(name),
        });

        return decompile_data(ctx, name, variant.data(), stream_file_name).await;
    }

    let Some(file_name) = variant.data_file_name() else {
        eyre::bail!("Texture file has no data and no data file");
    };

    tracing::debug!("Decompiling texture from external file '{}'", file_name);

    let path = match &ctx.game_dir {
        Some(dir) => dir.join("bundle").join(file_name),
        None => PathBuf::from("bundle").join(file_name),
    };

    tracing::trace!(path = %path.display());

    let data = fs::read(&path)
        .await
        .wrap_err_with(|| format!("Failed to read data file '{}'", path.display()))
        .with_suggestion(|| {
            "Provide a game directory in the config file or make sure the `data` directory is next to the provided bundle."
        })?;

    decompile_data(ctx, name, &data, None).await
}

#[tracing::instrument(skip(sjson, name), fields(sjson_len = sjson.as_ref().len(), name = %name.display()))]
pub async fn compile(
    name: IdString64,
    sjson: impl AsRef<str>,
    root: impl AsRef<Path> + std::fmt::Debug,
) -> Result<BundleFile> {
    let definitions: TextureDefinition = serde_sjson::from_str(sjson.as_ref())
        .wrap_err("Failed to deserialize SJSON")
        .with_section(|| sjson.as_ref().to_string().header("SJSON:"))?;

    let dds = {
        let path = root.as_ref().join(definitions.common.input.filename);
        fs::read(&path)
            .await
            .wrap_err_with(|| format!("Failed to read DDS file '{}'", path.display()))?
    };

    let (width, height) = {
        let mut r = Cursor::new(&dds);

        let magic = r.read_u32()?;
        eyre::ensure!(
            magic == 0x20534444,
            "Invalid magic bytes for DDS. Expected 0x20534444, got {:08x}",
            magic
        );

        r.seek(SeekFrom::Current(5))?;

        let width = r.read_u16()? as u32;
        let height = r.read_u16()? as u32;

        (width, height)
    };

    let mut w = Cursor::new(Vec::new());

    let texture = Texture {
        header: TextureHeader {
            // As long as we can't handle mipmaps, these two need be `0`
            flags: TextureFlags::empty(),
            n_streamable_mipmaps: 0,
            width,
            height,
        },
        data: dds,
        stream: None,
        category: IdString32::String(definitions.common.output.category),
    };
    texture.to_binary(&mut w)?;

    let mut variant = BundleFileVariant::new();
    variant.set_data(w.into_inner());

    let mut file = BundleFile::new(name, BundleFileType::Texture);
    file.add_variant(variant);

    Ok(file)
}
